/*----------------------------------------------------------------------
  File    : citrie.h
  Contents: management of a trie for condition independence test results
  Author  : Christian Borgelt
  History : 2006.03.05 file created
            2006.03.09 function cit_cicnt() added
            2007.01.28 function cit_stats() added
            2013.09.04 adapted to definition of ATTID
----------------------------------------------------------------------*/
#ifndef __CITRIE__
#define __CITRIE__
#include <stdio.h>
#include <float.h>
#include <math.h>
#include "memsys.h"

#ifdef _MSC_VER
#ifndef INFINITY
#define INFINITY     (DBL_MAX+DBL_MAX)
#endif
#ifndef NAN
#define NAN          (INFINITY-INFINITY)
#endif
#endif                          /* in case C99 is not supported */

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#ifdef EXTDATA

#ifndef ATTID
#define ATTID       ptrdiff_t   /* attribute identifier */
#endif

#else

#ifndef ATTID
#define ATTID       int         /* attribute identifier */
#endif

#endif
/*--------------------------------------------------------------------*/

#define int         1           /* to check definitions */
#define long        2           /* for certain types */
#define ptrdiff_t   3

/*--------------------------------------------------------------------*/

#if   ATTID==int
#ifndef ATTID_MAX
#define ATTID_MAX   INT_MAX     /* maximum attribute identifier */
#endif
#ifndef ATTID_FMT
#define ATTID_FMT   "d"         /* printf format code for int */
#endif
#ifndef aid_qsort
#define aid_qsort   int_qsort
#endif

#elif ATTID==long
#ifndef ATTID_MAX
#define ATTID_MAX   LONG_MAX    /* maximum attribute identifier */
#endif
#ifndef ATTID_FMT
#define ATTID_FMT   "ld"        /* printf format code for long */
#endif
#ifndef aid_qsort
#define aid_qsort   lng_qsort
#endif

#elif ATTID==ptrdiff_t
#ifndef ATTID_MAX
#define ATTID_MAX   PTRDIFF_MAX /* maximum attribute identifier */
#endif
#ifndef ATTID_FMT
#  ifdef _MSC_VER
#  define ATTID_FMT "Id"        /* printf format code for ptrdiff_t */
#  else
#  define ATTID_FMT "td"        /* printf format code for ptrdiff_t */
#  endif                        /* in case C99 is not supported */
#endif
#ifndef aid_qsort
#define aid_qsort   dif_qsort
#endif

#else
#error "ATTID must be either 'int', 'long' or 'ptrdiff_t'"
#endif

/*--------------------------------------------------------------------*/

#undef int                      /* remove preprocessor definitions */
#undef long                     /* needed for the type checking */
#undef ptrdiff_t

/*--------------------------------------------------------------------*/

#ifndef SIZE_FMT
#  ifdef _MSC_VER
#  define SIZE_FMT  "Iu"        /* printf format code for size_t */
#  else
#  define SIZE_FMT  "zu"        /* printf format code for size_t */
#  endif                        /* in case C99 is not supported */
#endif

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef struct cinode {         /* --- cond. indep. trie node --- */
  ATTID         id;             /* attribute identifier */
  double        val;            /* cond. dependence evaluation */
  struct cinode *succ;          /* next sibling */
  struct cinode *child;         /* first child node */
} CINODE;                       /* (cond. independence trie node) */

typedef struct {                /* --- cond. independence trie --- */
  MEMSYS        *msys;          /* memory management for nodes */
  ATTID         cnt;            /* number of attributes */
  ATTID         cimax;          /* maximum order of evaluations */
  size_t        cicnt;          /* number of stored evaluations */
  double        *vals;          /* evaluations for no conditions */
  ATTID         *buf;           /* buffer for condition identifiers */
  CINODE        *trees[1];      /* root nodes of the trees */
} CITRIE;                       /* (cond. independence trie) */

/*----------------------------------------------------------------------
  Functions
----------------------------------------------------------------------*/
extern CITRIE* cit_create (ATTID cnt);
extern void    cit_delete (CITRIE *cit);
extern ATTID   cit_cnt    (CITRIE *cit);
extern size_t  cit_cicnt  (CITRIE *cit);

extern int     cit_add    (CITRIE *cit, ATTID i, ATTID k,
                           ATTID *cons, ATTID n, double val);
extern double  cit_get    (CITRIE *cit, ATTID i, ATTID k,
                           ATTID *cons, ATTID n);
extern void    cit_desc   (CITRIE *cit, FILE *file);
extern void    cit_stats  (CITRIE *cit, FILE *file);

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define cit_cnt(t)        ((t)->cnt)
#define cit_cicnt(t)      ((t)->cicnt)
#define cit_cimax(t)      ((t)->cimax)

#endif
