/*----------------------------------------------------------------------
  File    : gramod.h
  Contents: graphical model management (based on directed graphs)
  Author  : Christian Borgelt
  History : 1996.02.23 file created as ascons.h from attset.h
            1996.04.09 function gm_rand modified (parameter randfn())
            1996.11.15 description modes added, gm_dmode extended
            1996.11.27 adapted to floating point counters
            1997.02.28 adapted to prob./poss. tree types
            1997.03.01 function gm_exec added
            1998.02.08 function gm_parse transferred from parser.h
            1998.03.01 adapted to extended prob./poss. tree structure
            1999.04.05 parameters 'attid' and 'conid' added to gm_check
            1999.04.30 definition of GM_ALL added
            2002.02.08 adapted to changed module ptree
            2002.03.11 parse flag GM_DFLT (create default network) added
            2002.03.30 function gm_evalx added (from sian.c)
            2002.04.04 functions gm_consum and gm_parsum added
            2002.04.11 function gm_tree added
            2002.07.16 parameter 'delnodes' added to function gm_clear
            2006.03.04 functions gm_tree and gm_citest added
            2006.03.10 log file added to induction functions
            2013.09.04 adapted to definitions ATTID, VALID, TPLID etc.
----------------------------------------------------------------------*/
#ifndef __GRAMOD__
#define __GRAMOD__
#if defined GM_PARSE && !defined PT_PARSE
#define PT_PARSE
#ifndef SCN_SCAN
#define SCN_SCAN
#endif
#endif
#include "scanner.h"
#if defined GM_EVAL || defined GM_INDUCE
#include "table.h"
#else
#include "attset.h"
#endif
#include "ptree.h"

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
/* --- types --- */
#define GM_PROB     PT_PROB     /* probabilistic (sum) */
#define GM_POSS     PT_POSS     /* possibilistic (maximum) */
#define GM_AUTO     PT_AUTO     /* determine type from input file */

/* --- description/parse modes --- */
#define GM_TITLE    PT_TITLE    /* print a title (as a comment) */
#define GM_INFO     PT_INFO     /* print add. info. (as a comment) */
#define GM_FULL     PT_FULL     /* print full trees (all branches) */
#define GM_REL      PT_REL      /* print relative frequencies */
#define GM_CDDS     0x0100      /* parse candidate lists */
#define GM_ALL      0x0200      /* parse a complete set of trees */
#define GM_DFLT     0x0400      /* create a default network */

/* --- search methods --- */
#define GM_NONE       0         /* no search, only parameter estim. */
#define GM_OWST       1         /* optimum weight spanning tree */
#define GM_EXTST      2         /* optimum spanning tree extension */
#define GM_TOPORD     3         /* cond. sel. on topological order */
#define GM_NOLOOP     4         /* cond. sel. without directed loops */
#define GM_UNRES      5         /* cond. sel. without restrictions */
#define GM_CITEST     6         /* cond. independence tests, directed */
#define GM_CIHEUR     7         /* ditto, without strict thinning */
#define GM_CIUNDIR    8         /* cond. independence tests, undir. */
#define GM_CIUDNIT    9         /* ditto, without intermed. thinning */
#define GM_SIAN      10         /* hypertree simulated annealing */
#define GM_UNKNOWN   11         /* unknown search method */

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef struct {                /* --- attribute specific data --- */
  ATT    *att;                  /* corresponding attribute */
  ATTID  cddsize;               /* size of candidate attribute vector */
  ATTID  cddcnt;                /* number of candidate attributes */
  ATTID  *cdds;                 /* candidate attributes */
  ATTID  depsize;               /* size of dependent attribute vector */
  ATTID  depcnt;                /* number of dependent attributes */
  ATTID  *deps;                 /* dependent attributes */
  PTREE  *ptree;                /* current prob./poss. tree */
  ATTID  pos;                   /* position in topological order */
  double prob;                  /* probability used in gm_execx() */
  double eval;                  /* evaluation result */
} GMATT;                        /* (attribute specific data) */

typedef struct {                /* --- graphical model --- */
  ATTSET *attset;               /* underlying attribute set */
  int    type;                  /* type of model (prob./poss.) */
  ATTID  attcnt;                /* number of attributes */
  ATTID  concnt;                /* number of conditions */
  double tplcnt;                /* number of tuples for norm. */
  WEIGHT lcorr;                 /* Laplace correction */
  ATTID  *order;                /* topological order of attributes */
  ATTID  *buf;                  /* buffer for execution and parsing */
  double eval;                  /* evaluation result */
  GMATT  atts[1];               /* vector of attribute specific data */
} GRAMOD;                       /* (attribute set conditions) */

/*----------------------------------------------------------------------
  Functions
----------------------------------------------------------------------*/
extern GRAMOD* gm_create  (ATTSET *attset, int type);
extern void    gm_delete  (GRAMOD *gm, int delas);

extern ATTSET* gm_attset  (GRAMOD *gm);
extern ATTID   gm_attcnt  (GRAMOD *gm);
extern int     gm_type    (GRAMOD *gm);
extern double  gm_setnorm (GRAMOD *gm, double tplcnt);
extern double  gm_getnorm (GRAMOD *gm);
extern WEIGHT  gm_setlc   (GRAMOD *gm, WEIGHT lcorr);
extern WEIGHT  gm_getlc   (GRAMOD *gm);
extern ATTID   gm_consum  (GRAMOD *gm);
extern ATTID   gm_parsum  (GRAMOD *gm);

extern ATTID   gm_cddcnt  (GRAMOD *gm, ATTID attid);
extern ATTID   gm_cddadd  (GRAMOD *gm, ATTID attid, ATTID cddid);
extern void    gm_cddrem  (GRAMOD *gm, ATTID attid, ATTID index);
extern ATTID   gm_cddid   (GRAMOD *gm, ATTID attid, ATTID index);

extern ATTID   gm_concnt  (GRAMOD *gm, ATTID attid);
extern ATTID   gm_conadd  (GRAMOD *gm, ATTID attid, ATTID conid);
extern void    gm_conrem  (GRAMOD *gm, ATTID attid, ATTID index);
extern ATTID   gm_conid   (GRAMOD *gm, ATTID attid, ATTID index);
extern ATT*    gm_con     (GRAMOD *gm, ATTID attid, ATTID index);

extern ATTID   gm_depcnt  (GRAMOD *gm, ATTID attid);
extern ATTID   gm_depid   (GRAMOD *gm, ATTID attid, ATTID index);
extern ATT*    gm_dep     (GRAMOD *gm, ATTID attid, ATTID index);

extern PTREE*  gm_ptree   (GRAMOD *gm, ATTID attid);
extern void    gm_clear   (GRAMOD *gm, int delnodes);
extern int     gm_aggr    (GRAMOD *gm, int distuv);
extern int     gm_check   (GRAMOD *gm, ATTID attid, ATTID conid);
extern int     gm_rand    (GRAMOD *gm, RANDFN randfn);
extern double  gm_exec    (GRAMOD *gm);
extern int     gm_execx   (GRAMOD *gm, double res[]);

#ifdef GM_EVAL
extern int     gm_estim   (GRAMOD *gm, TABLE *table, int distuv);
extern double  gm_eval    (GRAMOD *gm, int measure, double *params);
extern double  gm_evalx   (GRAMOD *gm, TABLE *table, double res[]);
#endif
#ifdef GM_INDUCE
extern CCHAR*  gm_sname   (int search);
extern double  gm_local   (GRAMOD *gm, int lwise, double minimp);
extern double  gm_params  (GRAMOD *gm, TABLE *table,
                           int measure, double *params);
extern double  gm_tree    (GRAMOD *gm, TABLE *table, int search,
                           int measure, double *params, double minval,
                           ATTID maxtest, FILE *log);
extern double  gm_citest  (GRAMOD *gm, TABLE *table, int search,
                           int measure, double *params, double thresh,
                           int lwise, double mrgimp, ATTID maxcon,
                           ATTID maxtest, FILE *log);
extern double  gm_consel  (GRAMOD *gm, TABLE *table, int search,
                           int measure, double *params, double minimp,
                           int lwise, double mrgimp, ATTID maxcon,
                           ATTID maxtest, FILE *log);
extern double  gm_sian    (GRAMOD *gm, TABLE *table, int trials,
                           ATTID maxsize, double keep, double frac,
                           double ppwgt, int distuv, RANDFN randfn);
#endif
extern int     gm_desc    (GRAMOD *gm, FILE *file, int mode,int maxlen);
#ifdef GM_PARSE
extern GRAMOD* gm_parse   (ATTSET *attset, SCANNER *scan, int type);
#endif

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define gm_attset(c)      ((c)->attset)
#define gm_attcnt(c)      ((c)->attcnt)
#define gm_type(c)        ((c)->type)
#define gm_getnorm(c)     ((c)->tplcnt)
#define gm_getlc(c)       ((c)->lcorr)

#define gm_cddcnt(c,a)    ((c)->atts[a].cddcnt)
#define gm_cddid(c,a,i)   ((c)->atts[a].cdds[i])

#define gm_concnt(c,a)    ((a) < 0) ? (c)->concnt \
                                    : pt_concnt((c)->atts[a].ptree)
#define gm_conid(c,a,i)   pt_conid((c)->atts[a].ptree, i)
#define gm_con(c,a,i)     pt_con((c)->atts[a].ptree, i)
#define gm_ptree(c,a)     ((c)->atts[a].ptree)

#define gm_depcnt(c,a)    ((c)->atts[a].depcnt)
#define gm_depid(c,a,i)   ((c)->atts[a].deps[i])
#define gm_dep(c,a,i)     (as_att((c)->attset, (c)->atts[a].deps[i]))

#endif
