/*----------------------------------------------------------------------
  File    : graph.h
  Contents: Management of mixed graphs for graphical model construction
  Author  : Christian Borgelt
  History : 2005.12.22 file created
            2006.02.01 function gr_transto added
            2006.03.02 some adjacent nodes functions added/redfined
            2006.03.03 parameter 'mode' added to function gr_common
            2006.03.09 extended functions added
            2006.03.11 function gr_dirpath added, GR_CHILDREN added
            2006.12.18 functions gr_undirect and gr_moralize added
            2013.09.04 adapted to definition of NODEID
----------------------------------------------------------------------*/
#ifndef __GRAPH__
#define __GRAPH__
#ifdef GR_PARSE
#ifndef SCN_SCAN
#define SCN_SCAN
#endif
#include "scanner.h"
#endif
#ifdef GR_EXTFN
#include "attset.h"
#ifndef NODEID
#define NODEID      ATTID
#endif
#endif

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#ifndef NODEID
#define NODEID      int         /* node identifier */
#endif

/*--------------------------------------------------------------------*/

#define int         1           /* to check definitions */
#define long        2           /* for certain types */
#define ptrdiff_t   3

/*--------------------------------------------------------------------*/

#if   NODEID==int
#ifndef NODEID_MAX
#define NODEID_MAX  INT_MAX     /* maximum node identifier */
#endif
#ifndef NODEID_FMT
#define NODEID_FMT  "d"         /* printf format code for int */
#endif
#ifndef nid_qsort
#define nid_qsort   int_qsort
#endif

#elif NODEID==long
#ifndef NODEID_MAX
#define NODEID_MAX  LONG_MAX    /* maximum node identifier */
#endif
#ifndef NODEID_FMT
#define NODEID_FMT  "ld"        /* printf format code for long */
#endif
#ifndef nid_qsort
#define nid_qsort   lng_qsort
#endif

#elif NODEID==ptrdiff_t
#ifndef NODEID_MAX
#define NODEID_MAX  PTRDIFF_MAX /* maximum node identifier */
#endif
#ifndef NODEID_FMT
#  ifdef _MSC_VER
#  define NODEID_FMT  "Id"      /* printf format code for ptrdiff_t */
#  else
#  define NODEID_FMT  "td"      /* printf format code for ptrdiff_t */
#  endif                        /* in case C99 is not supported */
#endif
#ifndef nid_qsort
#define nid_qsort   dif_qsort
#endif

#else
#error "NODEID must be either 'int', 'long' or 'ptrdiff_t'"
#endif

/*--------------------------------------------------------------------*/

#undef int                      /* remove preprocessor definitions */
#undef long                     /* needed for the type checking */
#undef ptrdiff_t

/*--------------------------------------------------------------------*/

#ifndef SIZE_FMT
#  ifdef _MSC_VER
#  define SIZE_FMT  "Iu"        /* printf format code for size_t */
#  else
#  define SIZE_FMT  "zu"        /* printf format code for size_t */
#  endif                        /* in case C99 is not supported */
#endif

/*--------------------------------------------------------------------*/
/* --- modes for gr_common, gr_adjsto, and gr_transto --- */
#define GR_NOPARENT    0x01     /* exclude parent nodes */
#define GR_NOCHILD     0x02     /* exclude child  nodes */
#define GR_CHILDREN    0x04     /* only pure child nodes */
/* GR_CHILDREN is available only for function gr_common. */

/* --- modes for gr_desc --- */
#define GR_NAMES       0        /* refer to nodes by their names */
#define GR_IDS         1        /* refer to nodes by their ids */
#define GR_TITLE       2        /* print a title */

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef struct {                /* --- a node --- */
  NODEID deg;                   /* degree (number of incident edges) */
  NODEID indeg;                 /* number of incoming edges */
  NODEID outdeg;                /* number of outgoing edges */
  NODEID mark;                  /* marker for internal use */
  NODEID pos;                   /* position in topological order */
  NODEID size;                  /* size of the vector of adj. nodes */
  NODEID adjs[1];               /* indices of adjacent nodes */
} NODE;                         /* (node) */

/* Note that indeg +outdeg need not equal deg, since some edges may */
/* have both (and thus actually no) direction. The number of these  */
/* edges is indeg +outdeg -deg. The first indeg nodes in the vector */
/* of adjacent nodes are parents (incoming edges), the last outdeg  */
/* nodes are children (outgoing edges). The children thus start at  */
/* the index deg -outdeg. */

typedef struct {                /* --- a graph --- */
  NODEID nodecnt;               /* number of nodes */
  NODEID edgecnt;               /* number of edges */
  NODEID dircnt;                /* number of directed edges */
  NODEID *topord;               /* topological order of nodes */
  #ifdef GR_EXTFN               /* if to use extended functions */
  ATTSET *attset;               /* underlying attribute set */
  #else                         /* if only standard functions */
  void   *attset;               /* dummy */
  #endif
  NODE   *nodes[1];             /* vector of nodes */
} GRAPH;                        /* (graph) */

/*----------------------------------------------------------------------
  Functions
----------------------------------------------------------------------*/
extern GRAPH*        gr_create   (NODEID nodecnt);
extern void          gr_delete   (GRAPH *graph);

extern void          gr_nodecnt  (GRAPH *graph);
extern void          gr_edgecnt  (GRAPH *graph);
extern void          gr_dircnt   (GRAPH *graph);
extern void          gr_undircnt (GRAPH *graph);
extern void          gr_bidircnt (GRAPH *graph);

extern int           gr_addedge  (GRAPH *graph, NODEID src, NODEID dst);
extern int           gr_adddir   (GRAPH *graph, NODEID src, NODEID dst);
extern int           gr_addundir (GRAPH *graph, NODEID src, NODEID dst);
extern int           gr_addbidir (GRAPH *graph, NODEID src, NODEID dst);
extern int           gr_remedge  (GRAPH *graph, NODEID src, NODEID dst);
extern int           gr_remdir   (GRAPH *graph, NODEID src, NODEID dst);
extern int           gr_remundir (GRAPH *graph, NODEID src, NODEID dst);
extern int           gr_rembidir (GRAPH *graph, NODEID src, NODEID dst);

extern int           gr_edge     (GRAPH *graph, NODEID src, NODEID dst);
extern int           gr_dirpath  (GRAPH *graph, NODEID src, NODEID dst);

extern NODEID        gr_degree   (GRAPH *graph, NODEID id);
extern NODEID        gr_indeg    (GRAPH *graph, NODEID id);
extern NODEID        gr_outdeg   (GRAPH *graph, NODEID id);
extern NODEID        gr_inoutdeg (GRAPH *graph, NODEID id);

extern NODEID        gr_adjcnt   (GRAPH *graph, NODEID id);
extern NODEID        gr_parcnt   (GRAPH *graph, NODEID id);
extern NODEID        gr_chdcnt   (GRAPH *graph, NODEID id);
extern NODEID        gr_sibcnt   (GRAPH *graph, NODEID id);

extern const NODEID* gr_adjs     (GRAPH *graph, NODEID id);
extern const NODEID* gr_incoming (GRAPH *graph, NODEID id);
extern const NODEID* gr_outgoing (GRAPH *graph, NODEID id);
extern const NODEID* gr_inout    (GRAPH *graph, NODEID id);

extern const NODEID* gr_parents  (GRAPH *graph, NODEID id);
extern const NODEID* gr_children (GRAPH *graph, NODEID id);
extern const NODEID* gr_siblings (GRAPH *graph, NODEID id);

extern NODEID        gr_common   (GRAPH *graph, NODEID src, NODEID dst,
                                  int mode, NODEID *adjs);
extern NODEID        gr_adjsto   (GRAPH *graph, NODEID src, NODEID dst,
                                  int mode, NODEID *adjs);
extern NODEID        gr_transto  (GRAPH *graph, NODEID src, NODEID dst,
                                  int mode, NODEID *adjs);
extern NODEID        gr_maxclq   (GRAPH *graph,
                                  NODEID *clique, NODEID cnt);

extern int           gr_isdir    (GRAPH *graph);
extern int           gr_isdag    (GRAPH *graph);
extern const NODEID* gr_topord   (GRAPH *graph);
extern NODEID        gr_topos    (GRAPH *graph, NODEID id);
extern NODEID        gr_direct   (GRAPH *graph, int full);
extern NODEID        gr_undirect (GRAPH *graph);
extern NODEID        gr_moralize (GRAPH *graph);

extern void          gr_desc     (GRAPH *graph, FILE *file, int mode,
                                  int maxlen);
#ifdef GR_PARSE
extern GRAPH*        gr_parse    (SCANNER *scan);
#endif
#ifdef GR_EXTFN
extern GRAPH*        gr_createx  (ATTSET *attset);
extern void          gr_deletex  (GRAPH *graph, int delas);
extern ATTSET*       gr_attset   (GRAPH *graph);
#ifdef GR_PARSE
extern GRAPH*        gr_parsex   (SCANNER *scan, ATTSET *attset);
#endif
#endif

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#ifdef GR_EXTFN
#define gr_attset(g)        ((g)->attset)
#endif
#define gr_nodecnt(g)       ((g)->nodecnt)
#define gr_edgecnt(g)       ((g)->edgecnt)
#define gr_dircnt(g)        ((g)->dircnt)
#define gr_undircnt(g)      ((g)->edgecnt -(g)->dircnt)
#define gr_bidircnt(g)      ((g)->edgecnt -(g)->dircnt)

#define gr_adddir(g,s,d)    gr_addedge(g, s, d)
#define gr_addbidir(g,s,d)  gr_addundir(g, s, d)
#define gr_remdir(g,s,d)    gr_remedge(g, s, d)
#define gr_rembidir(g,s,d)  gr_remundir(g, s, d)

#define gr_degree(g,i)      ((g)->nodes[i]->deg)
#define gr_indeg(g,i)       ((g)->nodes[i]->indeg)
#define gr_outdeg(g,i)      ((g)->nodes[i]->outdeg)

#define gr_adjcnt(g,i)      ((g)->nodes[i]->deg)
#define gr_sibcnt(g,i)      gr_bideg(g,i)

#define gr_adjs(g,i)        ((const NODEID*)(g)->nodes[i]->adjs)
#define gr_incoming(g,i)    ((const NODEID*)(g)->nodes[i]->adjs)
#define gr_inout(g,i)       gr_outgoing(g,i)

#define gr_parents(g,i)     ((const NODEID*)(g)->nodes[i]->adjs)
#define gr_siblings(g,i)    gr_outgoing(g,i)

#define gr_isdir(g)         (((g)->undir != 0)      ? 0 : -1)
#define gr_isdag(g)         ((gr_topord(g) == NULL) ? 0 : -1)
#define gr_topos(g, i)      ((g)->nodes[i]->pos)

#endif
