/*----------------------------------------------------------------------
  File    : ptree4.c
  Contents: probability/possibility tree management
            evaluation functions
  Author  : Christian Borgelt
  History : 1995.10.22 file created as ctree.c
            1995.11.03 first version of pt_eval() programmed
            1995.12.11 information gain (ratio) added
            1996.01.16 ln(K2 metric) replaced by log_2(K2 metric)/N
            1996.02.05 some evaluation measures added
            1996.02.07 evaluation measure 'PT_GINI' added
            1996.02.23 functions pt_mname() and pt_minfo() added
            1996.06.23 comp. of specificity gain (ratio) redesigned
            1996.11.15 sym. information/specificity gain ratio added
            1997.03.06 some evaluation measures added
            1997.03.07 functions chi2() and muts() improved
            1997.03.10 second symmetric info./spec. gain ratio added
            1998.02.15 marginalization moved to function pt_eval
            1998.02.20 Bayesian Dirichlet l.e. uniform metric added
            1998.02.24 bug in function woevid() fixed
            1999.02.23 changed to Bayes factor computation (K2,BDeu)
            1999.04.30 several assertions added
            1999.11.13 evaluation measure 'PT_WDIFF' added
            2001.05.26 computation of ln(\Gamma(n)) improved
            2002.02.02 major redesign completed
            2002.02.04 quadratic information measures added
            2002.03.05 bug in function ginit1() fixed
            2002.07.01 evaluation adapted to number of paths to a leaf
            2002.07.02 functions bdm() and bdmod() redesigned
            2002.07.04 function logGamma() made external
            2002.07.17 function spec() redesigned, nsp() -> ptree3.c
            2002.07.19 function sort() made safe w.r.t. links
            2003.01.17 normalization of possibilistic measures added
            2003.01.18 bug in computation of balanced spc. gain fixed
            2013.09.04 adapted to definitions ATTID, VALID, TPLID etc.
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <float.h>
#include <assert.h>
#include "gamma.h"
#include "ptree.h"
#ifdef STORAGE
#include "storage.h"
#endif

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#ifndef M_PI                    /* if not defined (no Posix source) */
#define M_PI   3.14159265358979323846   /* \pi   */
#define M_LN2  0.693147180559945309417  /* ln(2) */
#endif

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef int    INITFN (PTREE *pt);
typedef double EVALFN (PTREE *pt);
typedef double LEAFFN (PTREE *pt, PTLVL *lvl, PTLEAF *leaf);

typedef struct {                /* --- function table element --- */
  INITFN *initfn;               /* initialization function */
  EVALFN *evalfn;               /* evaluation function */
  LEAFFN *leaffn;               /* leaf evaluation function */
} FTE;                          /* (function table element) */

/*----------------------------------------------------------------------
  Auxiliary Functions
----------------------------------------------------------------------*/

static void leafsort (PTLVL *lvl, PTNODE *node)
{                               /* --- sort the leaves */
  VALID  i;                     /* loop variable */
  PTNODE *child;                /* to traverse the child nodes */

  assert(lvl && node);          /* check the function arguments */
  if (lvl->index < 0) {         /* if on the leaf level, */
    node->succ = lvl->list;     /* add the current leaf node */
    lvl->list = node; }         /* at the head of the level list */
  else {                        /* if at an inner node */
    for (i = (lvl++)->cnt; --i >= 0; ) {
      child = node->children[i];/* traverse the children */
      if (child && (child->parent == node) && (child->index == i))
        leafsort(lvl, child);   /* if the child node exists and */
    }                           /* it is a genuine child (no link), */
  }                             /* sort the nodes recursively */
}  /* leafsort() */

/*--------------------------------------------------------------------*/

static void sumsqr (PTREE *pt)
{                               /* --- sum squared frequencies */
  VALID  i;                     /* loop variable */
  PTLVL  *lvl;                  /* leaf level description */
  PTLEAF *leaf;                 /* to traverse the leaves */
  WEIGHT *c;                    /* to traverse the buffer/counters */
  double s_i, s_j, s_ij, s;     /* sums of squared frequencies */

  assert(pt && (pt->total > 0));/* check the function arguments */
  s_i = s_j = s_ij = 0;         /* initialize the sums */
  lvl = pt->levels +pt->concnt; /* get the leaf level */
  for (c = lvl->buf +(i = lvl->cnt); --i >= 0; ) {
    --c; s_i += *c * *c; }      /* sum squared marginals */
  for (leaf = (PTLEAF*)lvl->list; leaf; leaf = leaf->succ) {
    if (leaf->total <= 0) continue; /* traverse nonempty leaves */
    s_j += leaf->total *leaf->total;
    s = 0;                      /* traverse the counter vector */
    for (c = leaf->cnts +(i = lvl->cnt); --i >= 0; ) {
      --c; s += *c * *c; }      /* sum squared joint distrib. terms */
    s_ij += leaf->eval = s;     /* note and sum the leaf terms */
  }
  pt->res0[0] = s_i;            /* s_i  = sum_i  N_i^2  */
  pt->res0[1] = s_j;            /* s_j  = sum_j  N_j^2  */
  pt->res0[2] = s_ij;           /* s_ij = sum_ij N_ij^2 */
}  /* sumsqr() */

/*----------------------------------------------------------------------
  Information/Specificity Functions
----------------------------------------------------------------------*/

static int info (PTREE *pt)
{                               /* --- Shannon information init. */
  VALID  i;                     /* loop variable */
  PTLVL  *lvl;                  /* leaf level description */
  PTLEAF *leaf;                 /* to traverse the leaves */
  WEIGHT *c;                    /* to traverse the buffers/counters */
  double s_i, s_j, s_ij, n, t;  /* sums of entropy terms, buffers */

  assert(pt && (pt->total > 0));/* check the function arguments */
  s_ij = s_i = s_j = 0;         /* and initialize the sums */
  lvl = pt->levels +pt->concnt; /* traverse the leaf level buffer */
  for (c = lvl->buf +(i = lvl->cnt); --i >= 0; )
    if (*--c > 0) s_i += *c *log(*c);
  for (leaf = (PTLEAF*)lvl->list; leaf; leaf = leaf->succ) {
    if (leaf->total <= 0) continue; /* traverse nonempty leaves */
    s_j += leaf->total *log(leaf->total);
    t = 0;                      /* traverse the counter vector */
    for (c = leaf->cnts +(i = lvl->cnt); --i >= 0; )
      if (*--c > 0) t += *c *log(*c);
    s_ij += leaf->eval = t;     /* note and sum the leaf terms */
  }
  n = pt->total; t = n *log(n); /* compute N *log(N) only once */
  pt->res0[0] = t -s_i;         /* N H_i  = -N sum_i  p_i  *log(p_i)  */
  pt->res0[1] = t -s_j;         /* N H_j  = -N sum_j  p_j  *log(p_j)  */
  pt->res0[2] = t -s_ij;        /* N H_ij = -N sum_ij p_ij *log(p_ij) */
  pt->res0[3] = 1/(M_LN2 *n);   /* note normalization factors */
  pt->res0[4] = (pt->leafcnt <= 1) ? 0
              : 1/(n *log((double)pt->leafcnt));
  return 0;                     /* return 'ok' */
}  /* info() */

/*--------------------------------------------------------------------*/

static int quad (PTREE *pt)
{                               /* --- quadratic information init. */
  double t;                     /* squared total frequency */

  assert(pt && (pt->total > 0));/* check the function argument */
  sumsqr(pt);                   /* sum squares of frequencies */
  t = pt->total; t *= t;        /* compute N^2 only once */
  pt->res0[0] = t -pt->res0[0]; /* N^2/2 H^2_i  = N^2 -sum_i  N_i^2  */
  pt->res0[1] = t -pt->res0[1]; /* N^2/2 H^2_j  = N^2 -sum_j  N_j^2  */
  pt->res0[2] = t -pt->res0[2]; /* N^2/2 H^2_ij = N^2 -sum_ij N_ij^2 */
  pt->res0[3] = 2/t;            /* note normalization factors */
  pt->res0[4] = (pt->leafcnt <= 0) ? 0
              : 1/(t *(1 -1/(double)pt->leafcnt));
  return 0;                     /* return 'ok' */
}  /* quad() */

/*--------------------------------------------------------------------*/

static int spec (PTREE *pt)
{                               /* --- specificity measures init. */
  VALID  i;                     /* loop variable */
  PTLVL  *lvl;                  /* leaf level description */
  PTLEAF *leaf;                 /* to traverse the leaves */
  WEIGHT *b, *c;                /* to traverse the buffers/counters */

  assert(pt && (pt->total > 0));/* check the function argument */
  lvl = pt->levels +pt->concnt; /* get the leaf level and compute */
  pt->res0[0] = pt_nsp(lvl->buf, lvl->cnt);         /* nsp(child) */
  if (!pt->buf) {               /* if there is no evaluation buffer */
    pt->buf = (WEIGHT*)malloc((size_t)lvl->cnt
                             *(size_t)pt->leafcnt *sizeof(WEIGHT));
    if (!pt->buf) return -1;    /* allocate an evaluation buffer */
  }                             /* for the nonspecificity comps. */
  b = pt->buf;                  /* process distribution on conditions */
  for (leaf = (PTLEAF*)lvl->list; leaf; leaf = leaf->succ)
    if (leaf->total > 0) *b++ = (WEIGHT)leaf->total;
  pt->res0[1] = pt_nsp(pt->buf, (VALID)(b -pt->buf));
  b = pt->buf;                  /* process the joint distribution */
  for (leaf = (PTLEAF*)lvl->list; leaf; leaf = leaf->succ) {
    if (leaf->total <= 0) continue; /* traverse nonempty leaves */
    for (c = leaf->cnts +(i = lvl->cnt); --i >= 0; )
      if (*--c > 0) *b++ = *c;  /* copy nonzero counters to buffer */
  }                             /* and compute S_ij from them */
  pt->res0[2] = pt_nsp(pt->buf, (VALID)(b -pt->buf));
  pt->res0[3] = 1/M_LN2/((pt->tplcnt > 0) ? (double)pt->tplcnt : 1);
  pt->res0[4] = (pt->leafcnt > 1) ? M_LN2/log((double)pt->leafcnt) : 1;
  return 0;                     /* return 'ok' */
}  /* spec() */

/*--------------------------------------------------------------------*/

static double gain (PTREE *pt)
{ double *r = pt->res1;         /* --- information/specificity gain */
  return (r[0] +r[1] -r[2]) *r[3];
}  /* gain() */                 /* H_i +H_j -H_ij */

static double gbal (PTREE *pt)
{ double *r = pt->res1;         /* --- balanced info./spec. gain */
  return (r[0] +r[1] -r[2]) *r[3] *r[4];
}  /* gbal() */                 /* (H_i +H_j -H_ij) /r_j */

static double gratio (PTREE *pt)
{ double *r = pt->res1;         /* --- info./spec. gain ratio */
  return (r[1] > 0) ? (r[0] +r[1] -r[2]) /r[1] : 0;
}  /* gratio() */               /* (H_i +H_j -H_ij) /H_j */

static double sgr1 (PTREE *pt)
{ double *r = pt->res1;         /* --- sym. info./spec. gain ratio 1 */
  return (r[2] > 0) ? (r[0] +r[1] -r[2]) /r[2] : 0;
}  /* sgr1() */                 /* (H_i +H_j -H_ij) /H_ij */

static double sgr2 (PTREE *pt)
{ double *r = pt->res1;         /* --- sym. info./spec. gain ratio 2 */
  return (r[0] +r[1] > 0) ? (r[0] +r[1] -r[2]) /(r[0] +r[1]) : 0;
}  /* sgr2() */                 /* (H_i +H_j -H_ij) /(H_i +H_j) */

/*----------------------------------------------------------------------
  Gini Index and Related Measures
----------------------------------------------------------------------*/

static int ginit1 (PTREE *pt)
{                               /* --- Gini index initialization */
  PTLVL  *lvl;                  /* leaf level description */
  PTLEAF *leaf;                 /* to traverse the leaves */
  double w_ij;                  /* weighted sum of squared freqs. */

  assert(pt && (pt->total > 0));/* check the function arguments */
  sumsqr(pt); w_ij = 0;         /* compute sums of squared freqs. */
  lvl = pt->levels +pt->concnt; /* get the leaf level */
  for (leaf = (PTLEAF*)lvl->list; leaf; leaf = leaf->succ)
    if (leaf->total > 0) w_ij += leaf->eval /leaf->total;
  pt->res0[3] = w_ij;           /* w_ij = sum_j (1/N_j) sum_i N_ij^2 */
  return 0;                     /* return 'ok' */
}  /* ginit1() */

/*--------------------------------------------------------------------*/

static int ginit2 (PTREE *pt)
{                               /* --- symmetric Gini index init. */
  VALID  i;                     /* loop variable */
  PTLVL  *lvl;                  /* leaf level description */
  PTLEAF *leaf;                 /* to traverse the leaves */
  WEIGHT *b, *c;                /* to traverse the buffer/counters */
  double w_ij, w_ji;            /* weighted sums of squared freqs. */

  assert(pt && (pt->total > 0));/* check the function arguments */
  sumsqr(pt); w_ij = w_ji = 0;  /* compute sums of squared freqs. */
  lvl = pt->levels +pt->concnt; /* get the leaf level and */
  b   = lvl->buf +lvl->cnt;     /* the second evaluation buffer */
  for (leaf = (PTLEAF*)lvl->list; leaf; leaf = leaf->succ) {
    if (leaf->total <= 0) continue; /* traverse nonempty leaves */
    w_ij += leaf->eval / leaf->total;
    b    += lvl->cnt;           /* traverse the counter vector */
    for (c = leaf->cnts +(i = lvl->cnt); --i >= 0; ) {
      --c; *--b += *c * *c; }   /* sum the squared counters */
  }                             /* in the evaluation buffer */
  for (c = b +(i = lvl->cnt); --i >= 0; )
    w_ji += *--c / *--b;        /* weight and sum the terms */
  pt->res0[3] = w_ij;           /* w_ij = sum_j (1/N_j) sum_i N_ij^2 */
  pt->res0[4] = w_ji;           /* w_ji = sum_i (1/N_i) sum_j N_ij^2 */
  return 0;                     /* return 'ok' */
}  /* ginit2() */

/*--------------------------------------------------------------------*/

static double gini (PTREE *pt)
{ double *r = pt->res1;         /* --- normal Gini index */
  return (r[3] -r[0] /pt->total) /pt->total;
}  /* gini() */                 /* w_ij /N - s_i /N^2 */

static double ginisym (PTREE *pt)
{ double *r = pt->res1;         /* --- symmetric Gini index */
  double t  = 2 *pt->total *pt->total -r[0] -r[1];
  return (t > 0) ? ((r[3] +r[4]) *pt->total -r[0] -r[1]) /t : 0;
}  /* ginisym() */              /* (N w_ji +N w_ij -s_i -s_j) */
                                /* / (2 N^2 -s_i -s_j) */

static double ginimod (PTREE *pt)
{ double *r = pt->res1;         /* --- modified Gini index */
  return (r[1] > 0) ? (r[2] /r[1] -r[0] /pt->total) : 0;
}  /* ginimod() */              /* s_ij /s_j - s_i /N */

static double relief (PTREE *pt)
{ double *r = pt->res1;         /* --- relief measure */
  double t  = pt->total *pt->total -r[0];
  return ((r[0] > 0) && (t > 0)) ? (r[2] /r[0] -(r[1] -r[2]) /t) : 0;
}  /* relief() */               /* s_ij /s_i -(s_j -s_ij) /(N^2 -s_i) */

/*----------------------------------------------------------------------
  Bayesian Measures
----------------------------------------------------------------------*/

static int bdm (PTREE *pt)
{                               /* --- Bayesian-Dirichlet metric */
  VALID  i;                     /* loop variable */
  PTLVL  *lvl;                  /* leaf level description */
  PTLEAF *leaf;                 /* to traverse the leaves */
  WEIGHT *c;                    /* to traverse the buffer/counters */
  double p, q, s, a;            /* (sum of) prior(s), sensitivity */
  double bdm, t;                /* Bayesian-Dirichlet metric, buffer */
  int    x = 0;                 /* flag for multiple paths */

  assert(pt && (pt->total > 0));/* check the function arguments */
  lvl = pt->levels +pt->concnt; /* get the leaf level description */
  a   = (pt->params[0] > -1) ? pt->params[0]+1 : 1;
  a  *= a;                      /* get the sensitivity parameter */
  p   = (pt->params[1] != 0) ? pt->params[1]   : 1;
  if (p < 0) p /= (double)-lvl->cnt; /* get prior/equiv. sample size */
  for (t = 0, c = lvl->buf +(i = lvl->cnt); --i >= 0; )
    t += logGamma(*--c *a +p);  /* process the marginal distribution */
  s  = (double) lvl->cnt *p;    /* and compute the reference value */
  t += (double)-lvl->cnt *logGamma(p)
     + (logGamma(s) -logGamma(pt->total *a +s));
  pt->res0[0] = t;              /* note the reference result */
  if (pt->params[1] < 0) {      /* adapt the prior for the version */
    p /= (double)pt->fullcnt; x = 1; }    /* likelihood equivalent */
  bdm = 0;                      /* process the joint distribution */
  for (leaf = (PTLEAF*)lvl->list; leaf; leaf = leaf->succ) {
    if (leaf->total <= 0) continue;         /* traverse the */
    q = (x) ? p *(double)leaf->pathcnt : p; /* nonempty leaves */
    for (t = 0, c = leaf->cnts +(i = lvl->cnt); --i >= 0; )
      t += logGamma(*--c *a +q);/* process the cond. distribution */
    s  = (double) lvl->cnt *q;  /* and compute a leaf term */
    t += (double)-lvl->cnt *logGamma(q)
       + (logGamma(s) -logGamma(leaf->total *a +s));
    bdm += leaf->eval = t;      /* note the result in the leaf */
  }                             /* and sum the leaf terms */
  pt->res0[1] = bdm;            /* note the Bayesian-Dirichlet metric */
  return 0;                     /* return 'ok' */
}  /* bdm() */

/*--------------------------------------------------------------------*/

static int bdmod (PTREE *pt)
{                               /* --- modified BD / K2 metric */
  VALID  i;                     /* loop variable */
  PTLVL  *lvl;                  /* leaf level description */
  PTLEAF *leaf;                 /* to traverse the leaves */
  WEIGHT *c;                    /* to traverse the buffer/counters */
  double p, q, a;               /* prior and sensitivity */
  double bdm, s, t;             /* Bayesian-Dirichlet metric, buffers */
  int    x = 0;                 /* flag for multiple paths */

  assert(pt && (pt->total > 0));/* check the function arguments */
  lvl = pt->levels +pt->concnt; /* get the leaf level */
  a   = (pt->params[0] >= 0) ? pt->params[0] : 0;
  if (a < 0) a = 0;             /* get the sensitivity parameter */
  p   = (pt->params[1] != 0) ? pt->params[1] : 1;
  if (p < 0) p /= (double)-lvl->cnt; /* get prior/equiv. sample size */
  for (t = 0, c = lvl->buf +(i = lvl->cnt); --i >= 0; ) {
    s = *--c *a +p; t += logGamma(s +*c) -logGamma(s); }
  s  = pt->total*a +(double)lvl->cnt*p; /* process marginal dist. */
  t += logGamma(s) -logGamma(s +pt->total);
  pt->res0[0] = t;              /* note the reference result */
  if (pt->params[1] < 0) {      /* adapt the prior for the version */
    p /= (double)pt->fullcnt; x = 1; }    /* likelihood equivalent */
  bdm = 0;                      /* process the joint distribution */
  for (leaf = (PTLEAF*)lvl->list; leaf; leaf = leaf->succ) {
    if (leaf->total <= 0) continue; /* traverse the nonempty leaves */
    q = (x) ? p *(double)leaf->pathcnt : p; /* process cond. dists. */
    for (t = 0, c = leaf->cnts +(i = lvl->cnt); --i >= 0; ) {
      s = *--c *a +q; t += logGamma(s +*c) -logGamma(s); }
    s  = leaf->total *a +(double)lvl->cnt *q;
    t += logGamma(s) -logGamma(s +leaf->total);
    bdm += leaf->eval = t;      /* note the result in the leaf */
  }                             /* and sum the leaf terms */
  pt->res0[1] = bdm;            /* note the Bayesian-Dirichlet metric */
  return 0;                     /* return 'ok' */
}  /* bdmod() */

/*--------------------------------------------------------------------*/

static double bfactor (PTREE *pt)
{                               /* --- Bayes factor computation */
  return (pt->res1[1] -pt->res1[0]) /(M_LN2 *pt->total);
}  /* bfactor() */              /* compute log_2(bdm(G) /bdm(G_0)) */

/*----------------------------------------------------------------------
  Description Length Measures
----------------------------------------------------------------------*/

static int dlrel (PTREE *pt)
{                               /* --- desc. length (rel. freq.) */
  VALID  i;                     /* loop variable */
  PTLVL  *lvl;                  /* leaf level description */
  PTLEAF *leaf;                 /* to traverse the leaves */
  WEIGHT *c;                    /* to traverse the buffer/counters */
  double dat, mod, lGc, s;      /* description lengths, buffers */

  assert(pt && (pt->total > 0));/* check the function arguments */
  lvl = pt->levels +pt->concnt; /* note ln(gamma(number of values)) */
  lGc = logGamma((double)lvl->cnt);
  s   = 0;                      /* process the marginal distribution */
  for (c = lvl->buf +(i = lvl->cnt); --i >= 0; )
    if (*--c > 0) s += *c *log(*c);
  pt->res0[0] = pt->total *log(pt->total) -s;
  pt->res0[1] = logGamma(pt->total +(double)lvl->cnt)
              - logGamma(pt->total +1) -lGc;
  dat = mod = 0;                /* traverse the cond. distributions */
  for (leaf = (PTLEAF*)lvl->list; leaf; leaf = leaf->succ) {
    if (leaf->total <= 0) continue; /* traverse nonempty leaves */
    s = 0;                      /* process a conditional distribution */
    for (c = leaf->cnts +(i = lvl->cnt); --i >= 0; )
      if (*--c > 0) s += *c *log(*c);
    dat += leaf->eval = leaf->total *log(leaf->total) -s;
    mod += logGamma(leaf->total +(double)lvl->cnt)
         - logGamma(leaf->total +1) -lGc;
  }                             /* note and sum the results */
  pt->res0[2] = dat;            /* posterior data  coding length */
  pt->res0[3] = mod;            /* posterior model coding length */
  return 0;                     /* return 'ok' */
}  /* dlrel() */

/*--------------------------------------------------------------------*/

static int dlabs (PTREE *pt)
{                               /* --- desc. length (abs. freq.) */
  VALID  i;                     /* loop variable */
  PTLVL  *lvl;                  /* leaf level description */
  PTLEAF *leaf;                 /* to traverse the leaves */
  WEIGHT *b, *c;                /* to traverse the buffer/counters */
  double dat, mod, lGc, s, t;   /* description lengths, buffers */

  assert(pt && (pt->total > 0));/* check the function arguments */
  lvl = pt->levels +pt->concnt; /* compute ln(Gamma(num. values)) */
  lGc = logGamma((double)lvl->cnt);
  s   = 0;                      /* process the marginal distribution */
  for (b = lvl->buf +(i = lvl->cnt); --i >= 0; )
    s += logGamma(*--b +1);     /* compute bits for code book page */
  pt->res0[0] = (t = logGamma(pt->total +1)) -s;
  pt->res0[1] = logGamma(pt->total +(double)lvl->cnt) -t -lGc;
  dat = mod = 0;                /* process the cond. distributions */
  for (leaf = (PTLEAF*)lvl->list; leaf; leaf = leaf->succ) {
    if (leaf->total <= 0) continue; /* traverse the nonempty leaves */
    s = 0;                      /* process a conditional distribution */
    for (c = leaf->cnts +(i = lvl->cnt); --i >= 0; )
      s += logGamma(*--c +1);   /* compute bits for code book page */
    dat += leaf->eval = (t = logGamma(leaf->total +1)) -s;
    mod += logGamma(leaf->total +(double)lvl->cnt) -t -lGc;
  }                             /* note and sum the results */
  pt->res0[2] = dat;            /* data  description length */
  pt->res0[3] = mod;            /* model description length */
  return 0;                     /* return 'ok' */
}  /* dlabs() */

/*--------------------------------------------------------------------*/

static double rdlen (PTREE *pt)
{ double *r = pt->res1;         /* --- reduction of desc. length */
  double t = (pt->params[0]+1 > 0) ? pt->params[0]+1 : 1;
  return (r[0] -r[2] +(r[1] -r[3])/t) /(M_LN2 *pt->total);
}  /* rdlen() */                /* (data +model/(a+1)) / (ln(2) *N) */

/*--------------------------------------------------------------------*/

static int scinit (PTREE *pt)
{                               /* --- stochastic complexity */
  PTLVL  *lvl;                  /* leaf level description */
  PTLEAF *leaf;                 /* to traverse the leaves */
  double stc;                   /* temporary buffer */

  assert(pt && (pt->total > 0));/* check the function arguments */
  info(pt);                     /* initialize information measures */
  stc = log(0.5 *pt->total);    /* compute log(N/2) */
  lvl = pt->levels +pt->concnt; /* get the leaf level */
  for (leaf = (PTLEAF*)lvl->list; leaf; leaf = leaf->succ)
    if (leaf->total > 0) stc -= log(0.5 *leaf->total);
  pt->res0[3] = stc *0.5 *((double)lvl->cnt -1);
  pt->res0[4] = (1 -(double)pt->leafcnt)
              * log(pow(M_PI, 0.5 *(double)lvl->cnt))
                   - logGamma(0.5 *(double)lvl->cnt);
  return 0;                     /* compute the stochastic complexity */
}  /* scinit() */               /* and return 'ok' */

/*--------------------------------------------------------------------*/

static double stoco (PTREE *pt)
{ double *r = pt->res1;         /* --- stochastic complexity */
  return (r[0] +r[1] -r[2] +r[3] +r[4]) /(M_LN2 *pt->total);
}  /* stoco() */                /* information gain +penalty term */

/*----------------------------------------------------------------------
  Probabilistic Leaf Functions
----------------------------------------------------------------------*/

static double l_wdiff (PTREE *pt, PTLVL *lvl, PTLEAF *leaf)
{                               /* --- weighted differences */
  VALID  i;                     /* loop variable */
  WEIGHT *b, *c;                /* to traverse the buffer/counters */
  double sum = 0, t;            /* sum of difference terms, buffer */

  assert(pt && lvl && leaf);    /* check the function arguments */
  b = lvl->buf +(i = lvl->cnt); /* traverse marginal distribution */
  c = leaf->cnts +i;            /* and joint distribution */
  if      (pt->params[0] == 1) {
    while (--i >= 0) {          /* if the exponent is 1 */
      t = *--b *leaf->total - *--c *pt->total;
      sum += *c *fabs(t);       /* sum the weighted absolute */
    } }                         /* differences */
  else if (pt->params[0] == 2) {
    while (--i >= 0) {          /* if the exponent is 2 */
      t = *--b *leaf->total - *--c *pt->total;
      sum += *c *t *t;          /* sum the weighted squared */
    } }                         /* differences */
  else {                        /* if the exponent is anything */
    while (--i >= 0) {          /* other than 1 or 2 */
      t = *--b *leaf->total - *--c *pt->total;
      sum += *c *pow(fabs(t), pt->params[0]);
    }                           /* sum the weighted differences */
  }                             /* raised to the given power */
  return sum;                   /* return the computed sum */
}  /* l_wdiff() */

/*--------------------------------------------------------------------*/

static double l_chi2 (PTREE *pt, PTLVL *lvl, PTLEAF *leaf)
{                               /* --- chi^2 measure */
  VALID  i;                     /* loop variable */
  WEIGHT *b, *c;                /* to traverse the buffer/counters */
  double sum = 0, t, p;         /* sum of difference terms, buffer */

  assert(pt && lvl && leaf);    /* check the function arguments */
  b = lvl->buf +lvl->cnt;       /* traverse marginal and joint dist. */
  for (c = leaf->cnts +(i = lvl->cnt); --i >= 0; ) {
    --c; p = *--b *leaf->total; /* compute product of marginals */
    if (p <= 0) continue;       /* skip if product is zero */
    t = p - *c *pt->total;      /* compute difference to joint */
    sum += t *t /p;             /* sum the chi^2 terms */
  }                             /* (modified squared differences) */
  return sum;                   /* and return this sum */
}  /* l_chi2() */

/*--------------------------------------------------------------------*/

static double l_evid (PTREE *pt, PTLVL *lvl, PTLEAF *leaf)
{                               /* --- weight of evidence */
  VALID  i;                     /* loop variable */
  WEIGHT *b, *c;                /* to traverse the buffer/counters */
  double sum = 0, x, y, z;      /* sum of terms, buffers */

  assert(pt && lvl && leaf);    /* check the function arguments */
  b = lvl->buf +lvl->cnt;       /* traverse marginal and joint dist. */
  for (c = leaf->cnts +(i = lvl->cnt); --i >= 0; ) {
    z = (double)*--c * *--b;    /* compute common term only once */
    x = (double)*c *pt->total   -z; if (x == 0) continue;
    y = (double)*b *leaf->total -z; if (y == 0) continue;
    sum += *b *fabs(log(x/y));  /* sum the odds fractions weighted */
  }                             /* with the marginal probability */
  return sum *leaf->total;      /* and return this sum */
}  /* l_evid() */

/*--------------------------------------------------------------------*/

static double l_relev (PTREE *pt, PTLVL *lvl, PTLEAF *leaf)
{                               /* --- relevance */
  VALID  i;                     /* loop variable */
  WEIGHT *b, *c;                /* to traverse the buffer/counters */
  double sum = 0, max = 0, t;   /* sum/maximum of terms, buffers */

  assert(pt && lvl && leaf);    /* check the function arguments */
  b = lvl->buf +lvl->cnt;       /* traverse marginal and joint dist. */
  for (c = leaf->cnts +(i = lvl->cnt); --i >= 0; ) {
    --c; if (*--b <= 0) continue;   /* skip empty counters */
    t = (double)*c / *b;        /* compute frequency ratio */
    if (t <= max) sum += t;     /* and sum these ratios */
    else        { sum += max; max = t; }
  }                             /* exclude the/one maximal ratio */
  return sum;                   /* return the computed sum */
}  /* l_relev() */

/*--------------------------------------------------------------------*/

static double wdiff (PTREE *pt)
{ double t = pt->total;         /* --- weighted differences */
  return pt->res1[0] /pow(t, 2*pt->params[0] +1);
}  /* wdiff() */

static double chi2 (PTREE *pt)
{ double t = pt->total;         /* --- chi^2 measure */
  return pt->res1[0] /(t *t);
}  /* chi2() */

static double chi2nrm (PTREE *pt)
{ double t    = pt->total;      /* --- normalized chi^2 measure */
  PTLVL  *lvl = pt->levels +pt->concnt;
  if ((lvl->cnt < 2) || (pt->res1[1] <= 0)) return 0;
  return pt->res1[0] /(t *t *((double)lvl->cnt-1) *pt->res1[1]);
}  /* chi2nrm() */

static double evid (PTREE *pt)
{ double t = pt->total;         /* --- weight of evidence */
  return pt->res1[0] /(t *t *M_LN2);
}  /* evid() */

static double relev (PTREE *pt)
{                               /* --- relevance */
  return 1 -pt->res1[0] /(double)pt->levels[pt->concnt].cnt;
}  /* relev() */

/*----------------------------------------------------------------------
  Possibilistic Leaf Functions
----------------------------------------------------------------------*/

static double l_pdiff (PTREE *pt, PTLVL *lvl, PTLEAF *leaf)
{                               /* --- weighted differences */
  VALID  i;                     /* loop variable */
  WEIGHT *b, *c;                /* to traverse the buffer/counters */
  double sum = 0, t;            /* sum of difference terms, buffer */

  assert(pt && lvl && leaf);    /* check the function arguments */
  b = lvl->buf +(i = lvl->cnt); /* traverse marginal distribution */
  c = leaf->cnts +i;            /* and joint distribution */
  if      (pt->params[0] == 1) {
    while (--i >= 0) {          /* if the exponent is 1 */
      t = ((*--b < leaf->total) ? *b : leaf->total) - *--c;
      sum += *c *fabs(t);       /* sum the weighted absolute */
    } }                         /* differences */
  else if (pt->params[0] == 2) {
    while (--i >= 0) {          /* if the exponent is 2 */
      t = ((*--b < leaf->total) ? *b : leaf->total) - *--c;
      sum += *c *t *t;          /* sum the weighted squared */
    } }                         /* differences */
  else {                        /* if the exponent is anything */
    while (--i >= 0) {          /* other than 1 or 2 */
      t = ((*--b < leaf->total) ? *b : leaf->total) - *--c;
      sum += *c *pow(fabs(t), pt->params[0]);
    }                           /* sum the weighted differences */
  }                             /* raised to the given power */
  return sum;                   /* return the computed sum */
}  /* l_pdiff() */

/*--------------------------------------------------------------------*/

static double l_pchi2 (PTREE *pt, PTLVL *lvl, PTLEAF *leaf)
{                               /* --- possibilistic chi^2 measure */
  VALID  i;                     /* loop variable */
  WEIGHT *b, *c;                /* to traverse the buffer/counters */
  double sum = 0, t, p;         /* sum of difference terms, buffers */

  assert(pt && lvl && leaf);    /* check the function arguments */
  b = lvl->buf +lvl->cnt;       /* traverse marginal and joint dist. */
  for (c = leaf->cnts +(i = lvl->cnt); --i >= 0; ) {
    --c; p = (*--b < leaf->total) ? *b : leaf->total;
    if (p <= 0) continue;       /* compute minimum of marginals */
    t = p - *c;                 /* and difference to joint poss. */
    sum += t *t /p;             /* sum the chi^2 terms */
  }                             /* (modified squared differences) */
  return sum;                   /* and return this sum */
}  /* l_pchi2() */

/*--------------------------------------------------------------------*/

static double l_mutspc (PTREE *pt, PTLVL *lvl, PTLEAF *leaf)
{                               /* --- mutual specificity */
  VALID  i;                     /* loop variable */
  WEIGHT *b, *c;                /* to traverse the buffer/counters */
  double sum = 0, t;            /* sum of difference terms, buffer */

  assert(pt && lvl && leaf);    /* check the function arguments */
  b = lvl->buf +lvl->cnt;       /* traverse marginal and joint dist. */
  for (c = leaf->cnts +(i = lvl->cnt); --i >= 0; ) {
    t = (*--b < leaf->total) ? *b : leaf->total;
    if ((*--c > 0) && (t > 0)) sum += *c *log(t / *c);
  }                             /* sum weighted logs of quotients */
  return sum;                   /* and return this sum */
}  /* l_mutspc() */

/*--------------------------------------------------------------------*/

static double pdiff (PTREE *pt)
{                               /* --- weighted differences */
  return pt->res1[0]
       / ((pt->tplcnt > 0) ? pow(pt->tplcnt, pt->params[0]+1) : 1);
}  /* pdiff() */

static double pchi2 (PTREE *pt)
{                               /* --- poss. chi^2 measure */
  return pt->res1[0] /((pt->tplcnt > 0) ? pt->tplcnt : 1);
}  /* pchi2() */

static double mutspc (PTREE *pt)
{                               /* --- mutual specificity */
  return pt->res1[0] /((pt->tplcnt > 0) ? pt->tplcnt : 1);
}  /* mutspc() */

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
static int generic (PTREE *pt);

static FTE fn_prob[PT_UNKNOWN+1] = {
  /* PT_NONE     0 */ { (INITFN*)0, (EVALFN*)0, (LEAFFN*)0 },
  /* PT_INFGAIN  1 */ { info,       gain,       (LEAFFN*)0 },
  /* PT_INFGBAL  2 */ { info,       gbal,       (LEAFFN*)0 },
  /* PT_INFGR    3 */ { info,       gratio,     (LEAFFN*)0 },
  /* PT_INFSGR1  4 */ { info,       sgr1,       (LEAFFN*)0 },
  /* PT_INFSGR2  5 */ { info,       sgr2,       (LEAFFN*)0 },
  /* PT_QIGAIN   6 */ { quad,       gain,       (LEAFFN*)0 },
  /* PT_QIGBAL   7 */ { quad,       gbal,       (LEAFFN*)0 },
  /* PT_QIGR     8 */ { quad,       gratio,     (LEAFFN*)0 },
  /* PT_QISGR1   9 */ { quad,       sgr1,       (LEAFFN*)0 },
  /* PT_QISGR2  10 */ { quad,       sgr2,       (LEAFFN*)0 },
  /* PT_GINI    11 */ { ginit1,     gini,       (LEAFFN*)0 },
  /* PT_GINISYM 12 */ { ginit2,     ginisym,    (LEAFFN*)0 },
  /* PT_GINIMOD 13 */ { ginit1,     ginimod,    (LEAFFN*)0 },
  /* PT_RELIEF  14 */ { ginit1,     relief,     (LEAFFN*)0 },
  /* PT_WDIFF   15 */ { generic,    wdiff,      l_wdiff    },
  /* PT_CHI2    16 */ { generic,    chi2,       l_chi2     },
  /* PT_CHI2NRM 17 */ { generic,    chi2nrm,    l_chi2     },
  /* PT_WOEVID  18 */ { generic,    evid,       l_evid     },
  /* PT_RELEV   19 */ { generic,    relev,      l_relev    },
  /* PT_BDM     20 */ { bdm,        bfactor,    (LEAFFN*)0 },
  /* PT_BDMOD   21 */ { bdmod,      bfactor,    (LEAFFN*)0 },
  /* PT_RDLREL  22 */ { dlrel,      rdlen,      (LEAFFN*)0 },
  /* PT_RDLABS  23 */ { dlabs,      rdlen,      (LEAFFN*)0 },
  /* PT_STOCO   24 */ { scinit,     stoco,      (LEAFFN*)0 },
  /* PT_UNKNOWN 25 */ { (INITFN*)0, (EVALFN*)0, (LEAFFN*)0 },
};                              /* functions for probability trees */

static FTE fn_poss[PT_UNKNOWN+1] = {
  /* PT_NONE     0 */ { (INITFN*)0, (EVALFN*)0, (LEAFFN*)0 },
  /* PT_SPCGAIN  1 */ { spec,       gain,       (LEAFFN*)0 },
  /* PT_SPCGBAL  2 */ { spec,       gbal,       (LEAFFN*)0 },
  /* PT_SPCGR    3 */ { spec,       gratio,     (LEAFFN*)0 },
  /* PT_SPCSGR1  4 */ { spec,       sgr1,       (LEAFFN*)0 },
  /* PT_SPCGR2   5 */ { spec,       sgr2,       (LEAFFN*)0 },
  /*             6 */ { (INITFN*)0, (EVALFN*)0, (LEAFFN*)0 },
  /*             7 */ { (INITFN*)0, (EVALFN*)0, (LEAFFN*)0 },
  /*             8 */ { (INITFN*)0, (EVALFN*)0, (LEAFFN*)0 },
  /*             9 */ { (INITFN*)0, (EVALFN*)0, (LEAFFN*)0 },
  /*            10 */ { (INITFN*)0, (EVALFN*)0, (LEAFFN*)0 },
  /*            11 */ { (INITFN*)0, (EVALFN*)0, (LEAFFN*)0 },
  /*            12 */ { (INITFN*)0, (EVALFN*)0, (LEAFFN*)0 },
  /*            13 */ { (INITFN*)0, (EVALFN*)0, (LEAFFN*)0 },
  /*            14 */ { (INITFN*)0, (EVALFN*)0, (LEAFFN*)0 },
  /* PT_WDIFF   15 */ { generic,     pdiff,     l_pdiff    },
  /* PT_CHI2    16 */ { generic,     pchi2,     l_pchi2    },
  /*            17 */ { (INITFN*)0, (EVALFN*)0, (LEAFFN*)0 },
  /*            18 */ { (INITFN*)0, (EVALFN*)0, (LEAFFN*)0 },
  /*            19 */ { (INITFN*)0, (EVALFN*)0, (LEAFFN*)0 },
  /* PT_MUTSPC  20 */ { generic,     mutspc,    l_mutspc   },
  /*            21 */ { (INITFN*)0, (EVALFN*)0, (LEAFFN*)0 },
  /*            22 */ { (INITFN*)0, (EVALFN*)0, (LEAFFN*)0 },
  /*            23 */ { (INITFN*)0, (EVALFN*)0, (LEAFFN*)0 },
  /*            24 */ { (INITFN*)0, (EVALFN*)0, (LEAFFN*)0 },
  /*            25 */ { (INITFN*)0, (EVALFN*)0, (LEAFFN*)0 },
};                              /* functions for possibility trees */

/*----------------------------------------------------------------------
  Generic Initialization Function
----------------------------------------------------------------------*/

static int generic (PTREE *pt)
{                               /* --- generic init. function */
  PTLVL  *lvl;                  /* leaf level description */
  PTLEAF *leaf;                 /* to traverse the leaves */
  LEAFFN *fn;                   /* function to process a leaf */
  double sum = 0;               /* sum of leaf terms */

  assert(pt && (pt->total > 0));/* check the function arguments */
  fn  = ((pt->type == PT_PROB) ? fn_prob : fn_poss)[pt->measure].leaffn;
  lvl = pt->levels +pt->concnt; /* get the leaf level */
  for (leaf = (PTLEAF*)lvl->list; leaf; leaf = leaf->succ)
    sum += leaf->eval = fn(pt, lvl, leaf);
  pt->res0[0] = sum;            /* sum the difference terms */
  pt->res0[1] = (double)pt->leafcnt-1;  /* and note the sum */
  return 0;                     /* return 'ok' */
}  /* generic() */

/*----------------------------------------------------------------------
  Main Function
----------------------------------------------------------------------*/

double pt_eval (PTREE *pt, int measure, double *params)
{                               /* --- evaluate condition tree */
  VALID  i;                     /* loop variable */
  PTLVL  *lvl;                  /* to traverse the levels */
  PTLEAF *leaf;                 /* to traverse the leaves */
  WEIGHT *b, *c;                /* to traverse the buffers/counters */
  FTE    *fte;                  /* function table element */
  double r;                     /* evaluation result */

  assert(pt);                   /* check the function arguments */
  pt_total(pt);                 /* compute the counter totals and */
  pt->parcnt = 0;               /* the number of parameters/leaves */
  pt_parcnt(pt);                /* (and set the path counters) */
  if ((measure < PT_NONE) || (measure >= PT_UNKNOWN))
    return NAN;                 /* check the measure identifier */
  if ((measure == PT_NONE)      /* if no measure is selected */
  ||  (pt->concnt <= 0)         /* or there are no conditions */
  ||  (pt->total  <= 0))        /* or all counters are zero, */
    return 0;                   /* abort the function */
  pt->measure = measure;        /* note the evaluation measure */
  pt->params[0] = (params) ? params[0] : 0;  /* as well as the */
  pt->params[1] = (params) ? params[1] : 1;  /* parameters */
  lvl = pt->levels +pt->concnt; /* get the leaf level and */
  lvl->list = NULL;             /* sort the leaf nodes */
  leafsort(pt->levels, pt->levels->list);
  for (b = lvl->buf +(i = lvl->cnt); --i >= 0; )
    *--b = 0;                   /* clear the leaf marginals */
  for (leaf = (PTLEAF*)lvl->list; leaf; leaf = leaf->succ) {
    b += i = lvl->cnt;          /* traverse the leaf level and */
    c  = leaf->cnts +i;         /* the counters of each leaf */
    if (pt->type == PT_PROB) {  /* probability tree: sum */
      while (--i >= 0) *--b += *--c; }
    else {                      /* possibility tree: maximum */
      while (--i >= 0) if (*--c > *--b) *b = *c; }
  }                             /* compute marginal distribution */
  fte = ((pt->type == PT_PROB) ? fn_prob : fn_poss) +measure;
  if (fte->initfn(pt) != 0)     /* initialize the evaluation */
    return NAN;                 /* (may fail due to memory alloc.) */
  for (i = sizeof(pt->res0)/sizeof(*pt->res0); --i >= 0; )
    pt->res1[i] = pt->res0[i];  /* copy the intermediate results */
  r = fte->evalfn(pt);          /* evaluate the prob./poss. tree */
  return pt->eval = (fabs(r) < DBL_EPSILON) ? 0 : r;
}  /* pt_eval() */              /* return the evaluation result */

/*--------------------------------------------------------------------*/

double pt_reeval (PTREE *pt)
{                               /* --- reevaluate condition tree */
  FTE    *fte;                  /* function table element */
  double r;                     /* evaluation result */

  assert(pt);                   /* check the function arguments */
  if ((pt->measure == PT_NONE)  /* if no measure was selected */
  ||  (pt->concnt  <= 0)        /* or there are no conditions */
  ||  (pt->total   <= 0))       /* or all counters are zero, */
    return 0;                   /* abort the function */
  fte = ((pt->type == PT_PROB) ? fn_prob : fn_poss) +pt->measure;
  r   = fte->evalfn(pt);        /* reevaluate the prob./poss. tree */
  return pt->eval = (fabs(r) < DBL_EPSILON) ? 0 : r;
}  /* pt_reeval() */            /* return the evaluation result */
